@@ -13,13 +13,18 @@
 import copy
 import math
 
-from s3transfer.tasks import Task
-from s3transfer.tasks import SubmissionTask
-from s3transfer.tasks import CreateMultipartUploadTask
-from s3transfer.tasks import CompleteMultipartUploadTask
-from s3transfer.utils import get_callbacks
-from s3transfer.utils import calculate_range_parameter
-from s3transfer.utils import ChunksizeAdjuster
+from s3transfer.tasks import (
+    CompleteMultipartUploadTask,
+    CreateMultipartUploadTask,
+    SubmissionTask,
+    Task,
+)
+from s3transfer.utils import (
+    ChunksizeAdjuster,
+    calculate_range_parameter,
+    get_callbacks,
+    get_filtered_dict,
+)
 
 
 class CopySubmissionTask(SubmissionTask):
@@ -33,7 +38,8 @@ class CopySubmissionTask(SubmissionTask):
         'CopySourceSSECustomerKey': 'SSECustomerKey',
         'CopySourceSSECustomerAlgorithm': 'SSECustomerAlgorithm',
         'CopySourceSSECustomerKeyMD5': 'SSECustomerKeyMD5',
-        'RequestPayer': 'RequestPayer'
+        'RequestPayer': 'RequestPayer',
+        'ExpectedBucketOwner': 'ExpectedBucketOwner',
     }
 
     UPLOAD_PART_COPY_ARGS = [
@@ -48,6 +54,7 @@ class CopySubmissionTask(SubmissionTask):
         'SSECustomerAlgorithm',
         'SSECustomerKeyMD5',
         'RequestPayer',
+        'ExpectedBucketOwner',
     ]
 
     CREATE_MULTIPART_ARGS_BLACKLIST = [
@@ -58,11 +65,15 @@ class CopySubmissionTask(SubmissionTask):
         'CopySourceSSECustomerKey',
         'CopySourceSSECustomerAlgorithm',
         'CopySourceSSECustomerKeyMD5',
-        'MetadataDirective'
+        'MetadataDirective',
+        'TaggingDirective',
     ]
 
-    def _submit(self, client, config, osutil, request_executor,
-                transfer_future):
+    COMPLETE_MULTIPART_ARGS = ['RequestPayer', 'ExpectedBucketOwner']
+
+    def _submit(
+        self, client, config, osutil, request_executor, transfer_future
+    ):
         """
         :param client: The client associated with the transfer manager
 
@@ -89,9 +100,11 @@ class CopySubmissionTask(SubmissionTask):
             # of the client, they may have to provide the file size themselves
             # with a completely new client.
             call_args = transfer_future.meta.call_args
-            head_object_request = \
+            head_object_request = (
                 self._get_head_object_request_from_copy_source(
-                    call_args.copy_source)
+                    call_args.copy_source
+                )
+            )
             extra_args = call_args.extra_args
 
             # Map any values that may be used in the head object that is
@@ -99,24 +112,30 @@ class CopySubmissionTask(SubmissionTask):
             for param, value in extra_args.items():
                 if param in self.EXTRA_ARGS_TO_HEAD_ARGS_MAPPING:
                     head_object_request[
-                        self.EXTRA_ARGS_TO_HEAD_ARGS_MAPPING[param]] = value
+                        self.EXTRA_ARGS_TO_HEAD_ARGS_MAPPING[param]
+                    ] = value
 
             response = call_args.source_client.head_object(
-                **head_object_request)
+                **head_object_request
+            )
             transfer_future.meta.provide_transfer_size(
-                response['ContentLength'])
+                response['ContentLength']
+            )
 
         # If it is greater than threshold do a multipart copy, otherwise
         # do a regular copy object.
         if transfer_future.meta.size < config.multipart_threshold:
             self._submit_copy_request(
-                client, config, osutil, request_executor, transfer_future)
+                client, config, osutil, request_executor, transfer_future
+            )
         else:
             self._submit_multipart_request(
-                client, config, osutil, request_executor, transfer_future)
+                client, config, osutil, request_executor, transfer_future
+            )
 
-    def _submit_copy_request(self, client, config, osutil, request_executor,
-                             transfer_future):
+    def _submit_copy_request(
+        self, client, config, osutil, request_executor, transfer_future
+    ):
         call_args = transfer_future.meta.call_args
 
         # Get the needed progress callbacks for the task
@@ -134,14 +153,15 @@ class CopySubmissionTask(SubmissionTask):
                     'key': call_args.key,
                     'extra_args': call_args.extra_args,
                     'callbacks': progress_callbacks,
-                    'size': transfer_future.meta.size
+                    'size': transfer_future.meta.size,
                 },
-                is_final=True
-            )
+                is_final=True,
+            ),
         )
 
-    def _submit_multipart_request(self, client, config, osutil,
-                                  request_executor, transfer_future):
+    def _submit_multipart_request(
+        self, client, config, osutil, request_executor, transfer_future
+    ):
         call_args = transfer_future.meta.call_args
 
         # Submit the request to create a multipart upload and make sure it
@@ -160,8 +180,8 @@ class CopySubmissionTask(SubmissionTask):
                     'bucket': call_args.bucket,
                     'key': call_args.key,
                     'extra_args': create_multipart_extra_args,
-                }
-            )
+                },
+            ),
         )
 
         # Determine how many parts are needed based on filesize and
@@ -169,9 +189,11 @@ class CopySubmissionTask(SubmissionTask):
         part_size = config.multipart_chunksize
         adjuster = ChunksizeAdjuster()
         part_size = adjuster.adjust_chunksize(
-            part_size, transfer_future.meta.size)
+            part_size, transfer_future.meta.size
+        )
         num_parts = int(
-            math.ceil(transfer_future.meta.size / float(part_size)))
+            math.ceil(transfer_future.meta.size / float(part_size))
+        )
 
         # Submit requests to upload the parts of the file.
         part_futures = []
@@ -179,17 +201,27 @@ class CopySubmissionTask(SubmissionTask):
 
         for part_number in range(1, num_parts + 1):
             extra_part_args = self._extra_upload_part_args(
-                call_args.extra_args)
+                call_args.extra_args
+            )
             # The part number for upload part starts at 1 while the
             # range parameter starts at zero, so just subtract 1 off of
             # the part number
             extra_part_args['CopySourceRange'] = calculate_range_parameter(
-                part_size, part_number-1, num_parts, transfer_future.meta.size)
+                part_size,
+                part_number - 1,
+                num_parts,
+                transfer_future.meta.size,
+            )
             # Get the size of the part copy as well for the progress
             # callbacks.
             size = self._get_transfer_size(
-                part_size, part_number-1, num_parts, transfer_future.meta.size
+                part_size,
+                part_number - 1,
+                num_parts,
+                transfer_future.meta.size,
             )
+            # Get the checksum algorithm of the multipart request.
+            checksum_algorithm = call_args.extra_args.get("ChecksumAlgorithm")
             part_futures.append(
                 self._transfer_coordinator.submit(
                     request_executor,
@@ -203,15 +235,19 @@ class CopySubmissionTask(SubmissionTask):
                             'part_number': part_number,
                             'extra_args': extra_part_args,
                             'callbacks': progress_callbacks,
-                            'size': size
+                            'size': size,
+                            'checksum_algorithm': checksum_algorithm,
                         },
                         pending_main_kwargs={
                             'upload_id': create_multipart_future
-                        }
-                    )
+                        },
+                    ),
                 )
             )
 
+        complete_multipart_extra_args = self._extra_complete_multipart_args(
+            call_args.extra_args
+        )
         # Submit the request to complete the multipart upload.
         self._transfer_coordinator.submit(
             request_executor,
@@ -220,14 +256,15 @@ class CopySubmissionTask(SubmissionTask):
                 main_kwargs={
                     'client': client,
                     'bucket': call_args.bucket,
-                    'key': call_args.key
+                    'key': call_args.key,
+                    'extra_args': complete_multipart_extra_args,
                 },
                 pending_main_kwargs={
                     'upload_id': create_multipart_future,
-                    'parts': part_futures
+                    'parts': part_futures,
                 },
-                is_final=True
-            )
+                is_final=True,
+            ),
         )
 
     def _get_head_object_request_from_copy_source(self, copy_source):
@@ -237,21 +274,20 @@ class CopySubmissionTask(SubmissionTask):
             raise TypeError(
                 'Expecting dictionary formatted: '
                 '{"Bucket": bucket_name, "Key": key} '
-                'but got %s or type %s.'
-                % (copy_source, type(copy_source))
+                'but got %s or type %s.' % (copy_source, type(copy_source))
             )
 
     def _extra_upload_part_args(self, extra_args):
         # Only the args in COPY_PART_ARGS actually need to be passed
         # onto the upload_part_copy calls.
-        upload_parts_args = {}
-        for key, value in extra_args.items():
-            if key in self.UPLOAD_PART_COPY_ARGS:
-                upload_parts_args[key] = value
-        return upload_parts_args
+        return get_filtered_dict(extra_args, self.UPLOAD_PART_COPY_ARGS)
 
-    def _get_transfer_size(self, part_size, part_index, num_parts,
-                           total_transfer_size):
+    def _extra_complete_multipart_args(self, extra_args):
+        return get_filtered_dict(extra_args, self.COMPLETE_MULTIPART_ARGS)
+
+    def _get_transfer_size(
+        self, part_size, part_index, num_parts, total_transfer_size
+    ):
         if part_index == num_parts - 1:
             # The last part may be different in size then the rest of the
             # parts.
@@ -261,8 +297,10 @@ class CopySubmissionTask(SubmissionTask):
 
 class CopyObjectTask(Task):
     """Task to do a nonmultipart copy"""
-    def _main(self, client, copy_source, bucket, key, extra_args, callbacks,
-              size):
+
+    def _main(
+        self, client, copy_source, bucket, key, extra_args, callbacks, size
+    ):
         """
         :param client: The client to use when calling PutObject
         :param copy_source: The CopySource parameter to use
@@ -276,15 +314,28 @@ class CopyObjectTask(Task):
 
         """
         client.copy_object(
-            CopySource=copy_source, Bucket=bucket, Key=key, **extra_args)
+            CopySource=copy_source, Bucket=bucket, Key=key, **extra_args
+        )
         for callback in callbacks:
             callback(bytes_transferred=size)
 
 
 class CopyPartTask(Task):
     """Task to upload a part in a multipart copy"""
-    def _main(self, client, copy_source, bucket, key, upload_id, part_number,
-              extra_args, callbacks, size):
+
+    def _main(
+        self,
+        client,
+        copy_source,
+        bucket,
+        key,
+        upload_id,
+        part_number,
+        extra_args,
+        callbacks,
+        size,
+        checksum_algorithm=None,
+    ):
         """
         :param client: The client to use when calling PutObject
         :param copy_source: The CopySource parameter to use
@@ -298,6 +349,8 @@ class CopyPartTask(Task):
         :param callbacks: List of callbacks to call after copy part
         :param size: The size of the transfer. This value is passed into
             the callbacks
+        :param checksum_algorithm: The algorithm that was used to create the multipart
+            upload
 
         :rtype: dict
         :returns: A dictionary representing a part::
@@ -305,12 +358,25 @@ class CopyPartTask(Task):
             {'Etag': etag_value, 'PartNumber': part_number}
 
             This value can be appended to a list to be used to complete
-            the multipart upload.
+            the multipart upload. If a checksum is in the response,
+            it will also be included.
         """
         response = client.upload_part_copy(
-            CopySource=copy_source, Bucket=bucket, Key=key,
-            UploadId=upload_id, PartNumber=part_number, **extra_args)
+            CopySource=copy_source,
+            Bucket=bucket,
+            Key=key,
+            UploadId=upload_id,
+            PartNumber=part_number,
+            **extra_args,
+        )
         for callback in callbacks:
             callback(bytes_transferred=size)
         etag = response['CopyPartResult']['ETag']
-        return {'ETag': etag, 'PartNumber': part_number}
+        part_metadata = {'ETag': etag, 'PartNumber': part_number}
+        if checksum_algorithm:
+            checksum_member = f'Checksum{checksum_algorithm.upper()}'
+            if checksum_member in response['CopyPartResult']:
+                part_metadata[checksum_member] = response['CopyPartResult'][
+                    checksum_member
+                ]
+        return part_metadata
